# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Pydoc-markdown is an extensible framework for generating API documentation,
with a focus on Python source code and the Markdown output format.
"""

import dataclasses
import logging
import os
from pathlib import Path
import subprocess
import typing as t

import databind.core.annotations as A
import databind.json
import docspec
import tomli

from pydoc_markdown.interfaces import Context, Loader, Processor, Renderer, Resolver, Builder
from pydoc_markdown.contrib.loaders.python import PythonLoader
from pydoc_markdown.contrib.processors.filter import FilterProcessor
from pydoc_markdown.contrib.processors.crossref import CrossrefProcessor
from pydoc_markdown.contrib.processors.smart import SmartProcessor
from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer
from pydoc_markdown.util import ytemplate


__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '4.6.3'

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Hooks:
  pre_render: t.List[str] = dataclasses.field(default_factory=list, metadata={'alias': 'pre-render'})
  post_render: t.List[str] = dataclasses.field(default_factory=list, metadata={'alias': 'post-render'})


@dataclasses.dataclass
class PydocMarkdown:
  """
  This object represents the main configuration for Pydoc-Markdown.
  """

  #: A list of loader implementations that load #docspec.Module#s.
  #: Defaults to #PythonLoader.
  loaders: t.List[Loader] = dataclasses.field(default_factory=lambda: [PythonLoader()])

  #: A list of processor implementations that modify #docspec.Module#s. Defaults
  #: to #FilterProcessor, #SmartProcessor and #CrossrefProcessor.
  processors: t.List[Processor] = dataclasses.field(default_factory=lambda: [
    FilterProcessor(), SmartProcessor(), CrossrefProcessor()])

  #: A renderer for #docspec.Module#s. Defaults to #MarkdownRenderer.
  renderer: Renderer = dataclasses.field(default_factory=MarkdownRenderer)

  #: Hooks that can be executed at certain points in the pipeline. The commands
  #: are executed with the current `SHELL`.
  hooks: Hooks = dataclasses.field(default_factory=Hooks)

  # Hidden fields are filled at a later point in time and are not (de-) serialized.
  unknown_fields: t.List[str] = dataclasses.field(default_factory=list)

  def __post_init__(self) -> None:
    self.resolver: t.Optional[Resolver] = None
    self._context: t.Optional[Context] = None

  def load_config(self, arg: t.Union[str, dict]) -> None:
    """
    Loads the configuration from a nested data structure or filename as specified per the *data*
    argument. If a filename is specified, it may be a JSON, YAML or TOML file. If the name of the
    TOML file is `pyproject.yoml`, the configuration will be read from the `[tool.pydoc-markdown]`
    section.

    :param data: A nested structure or the path to a configuration file.
    """

    filename = None
    if isinstance(arg, str):
      filename = arg
      logger.info('Loading configuration file "%s".', filename)
      if filename.endswith('.toml'):
        data = tomli.loads(Path(filename).read_text())
      else:
        data = ytemplate.load(filename, {'env': ytemplate.Attributor(os.environ)})
      if filename == 'pyproject.toml':
        data = data['tool']['pydoc-markdown']
    else:
      data = arg

    unknown_keys = A.collect_unknowns()
    result = databind.json.new_mapper().deserialize(
      data,
      type(self),
      filename=filename,
      settings=[unknown_keys()])
    vars(self).update(vars(result))

    for loc, keys in unknown_keys:
      for key in keys:
        self.unknown_fields.append(str(loc.push_unknown(key).format()))

  def init(self, context: Context) -> None:
    """
    Initialize all plugins with the specified *context*. Cannot be called multiple times.
    If omitted, the plugins will be initialized with a default context before the load,
    process or render phase.
    """

    if self._context:
      raise RuntimeError('already initialized')
    self._context = context
    logger.debug('Initializing plugins with context %r', context)
    for loader in self.loaders:
      loader.init(context)
    for processor in self.processors:
      processor.init(context)
    self.renderer.init(context)

  def ensure_initialized(self) -> None:
    if not self._context:
      self.init(Context(directory='.'))

  def load_modules(self) -> t.List[docspec.Module]:
    """
    Loads modules via the #loaders.
    """

    logger.info('Loading modules.')
    self.ensure_initialized()
    modules: t.List[docspec.Module] = []
    for loader in self.loaders:
      modules.extend(loader.load())
    return modules

  def process(self, modules: t.List[docspec.Module]) -> None:
    """
    Process modules via the #processors.
    """

    self.ensure_initialized()
    if self.resolver is None:
      self.resolver = self.renderer.get_resolver(modules)
    for processor in self.processors:
      processor.process(modules, self.resolver)

  def render(self, modules: t.List[docspec.Module], run_hooks: bool = True) -> None:
    """
    Render modules via the #renderer.
    """

    self.ensure_initialized()
    if run_hooks:
      self.run_hooks('pre-render')
    if self.resolver is None:
      self.resolver = self.renderer.get_resolver(modules)
    self.renderer.process(modules, self.resolver)
    self.renderer.render(modules)
    if run_hooks:
      self.run_hooks('post-render')

  def build(self, site_dir: str) -> None:
    if not isinstance(self.renderer, Builder):
      name = type(self.renderer).__name__
      raise NotImplementedError('Renderer "{}" does not support building'.format(name))
    self.ensure_initialized()
    self.renderer.build(site_dir)

  def run_hooks(self, hook_name: str) -> None:
    assert self._context is not None

    # Remove the __PYVENV_LAUNCHER__ environment variable. This is needed if you are in a virtualenv and the hook
    # tries to invoke a script installed into a _different_ virtualenv. Otherwise, that script's execution of the
    # Python `site` module will set the `sys.prefix` the prefix of your terminal's activated virtualenv. The prefix
    # is then used to find site-packages, and thus none of the site-packages from the script's actual prefix are
    # detected.
    env = os.environ.copy()
    env.pop('__PYVENV_LAUNCHER__', None)

    for command in getattr(self.hooks, hook_name.replace('-', '_')):
      subprocess.check_call(command, shell=True, cwd=self._context.directory, env=env)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import dataclasses
import html
import io
import sys
import typing as t
from pathlib import Path

import docspec
from docspec_python import format_arglist
from yapf.yapflib.yapf_api import FormatCode  # type: ignore[import]

from pydoc_markdown.interfaces import Context, Renderer, Resolver, ResolverV2, SingleObjectRenderer, SinglePageRenderer, SourceLinker
from pydoc_markdown.util.docspec import format_function_signature, is_method, ApiSuite


def dotted_name(obj: docspec.ApiObject) -> str:
  return '.'.join(x.name for x in obj.path)


@dataclasses.dataclass
class MarkdownRenderer(Renderer, SinglePageRenderer, SingleObjectRenderer):
  """
  Produces Markdown files. This renderer is often used by other renderers, such as
  #MkdocsRenderer and #HugoRenderer. It provides a wide variety of options to customize
  the generated Markdown files.

  ### Options
  """

  #: The name of the file to render to. If no file is specified, it will
  #: render to stdout.
  filename: t.Optional[str] = None

  #: The encoding of the output file. This is ignored when rendering to
  #: stdout.
  encoding: str = 'utf-8'

  #: If enabled, inserts anchors before Markdown headers to ensure that
  #: links to the header work. This is enabled by default.
  insert_header_anchors: bool = True

  #: Generate HTML headers instead of Mearkdown headers. This is disabled
  #: by default.
  html_headers: bool = False

  #: Render names in headers as code (using backticks or `<code>` tags,
  #: depending on #html_headers). This is enabled by default.
  code_headers: bool = False

  #: Generate descriptive class titles by adding the word "Objects" if set to `True`. Otherwise,
  #: it can be a string that is appended or prepended (appended if the string begins with `$`).
  #: the class name. This is enabled by default.
  descriptive_class_title: t.Union[bool, str] = True

  #: Generate descriptivie module titles by adding the word "Module" before
  #: the module name. This is enabled by default.
  descriptive_module_title: bool = False

  #: Add the module name as a prefix to class & method names. This module name is
  #: also rendered as code if #code_headers is enabled. This is enabled
  #: by default.
  add_module_prefix: bool = True

  #: Add the class name as a prefix to method names. This class name is
  #: also rendered as code if #code_headers is enabled. This is enabled
  #: by default.
  add_method_class_prefix: bool = False

  #: Add the class name as a prefix to member names. This is enabled by
  #: default.
  add_member_class_prefix: bool = False

  #: Add the full module name as a prefix to the title of the header.
  #: This is disabled by default.
  add_full_prefix: bool = False

  #: If #add_full_prefix is enabled, this will result in the prefix to
  #: be wrapped in a `<sub>` tag.
  sub_prefix: bool = False

  #: Render the definition of data members as a code block. This is disabled
  #: by default.
  data_code_block: bool = False

  #: Max length of expressions. If this limit is exceeded, the remaining
  #: characters will be replaced with three dots. This is set to 100 by
  #: default.
  data_expression_maxlength: int = 100

  #: Render the class signature as a code block. This includes the "class"
  #: keyword, the class name and its bases. This is enabled by default.
  classdef_code_block: bool = True

  #: Render decorators before class definitions.
  classdef_with_decorators: bool = True

  #: Render classdef and function signature blocks in the Python help()
  #: style.
  signature_python_help_style: bool = False

  #: Render the function signature as a code block. This includes the "def"
  #: keyword, the function name and its arguments. This is enabled by
  #: default.
  signature_code_block: bool = True

  #: Render the function signature in the header. This is disabled by default.
  signature_in_header: bool = False

  #: Render the vertical bar '|' before function signature. This is enabled by default.
  signature_with_vertical_bar: bool = False

  #: Include the "def" keyword in the function signature. This is enabled
  #: by default.
  signature_with_def: bool = True

  #: Render the class name in the code block for function signature. Note
  #: that this results in invalid Python syntax to be rendered. This is
  #: disabled by default.
  signature_class_prefix: bool = False

  #: Render decorators before function definitions.
  signature_with_decorators: bool = True

  #: Render type hints for data elements in the header.
  render_typehint_in_data_header: bool = False

  #: Add the string "python" after the backticks for code blocks. This is
  #: enabled by default.
  code_lang: bool = True

  #: Render title of page at the beginning of the file.
  render_page_title: bool = False

  #: Render a table of contents at the beginning of the file.
  render_toc: bool = False

  #: The title of the "Table of Contents" header.
  render_toc_title: str = 'Table of Contents'

  #: The maximum depth of the table of contents. Defaults to 2.
  toc_maxdepth: int = 2

  #: Render module headers. This is enabled by default.
  render_module_header: bool = True

  #: Custom template for module header.
  render_module_header_template: str = ''

  #: Render docstrings as blockquotes. This is disabled by default.
  docstrings_as_blockquote: bool = False

  #: Use a fixed header level for every kind of API object. The individual
  #: levels can be defined with #header_level_by_type.
  use_fixed_header_levels: bool = True

  #: Fixed header levels by API object type.
  header_level_by_type: t.Dict[str, int] = dataclasses.field(default_factory=lambda: {
    'Module': 1,
    'Class': 2,
    'Method': 4,
    'Function': 4,
    'Variable': 4,
  })

  #: A plugin that implements the #SourceLinker interface to provide links to the
  #: source code of API objects. If this field is specified, the renderer will
  #: place links to the source code in the generated Markdown files.
  source_linker: t.Optional[SourceLinker] = None

  #: Allows you to define the position of the "view source" link in the Markdown
  #: file if a #source_linker is configured.
  # TODO: Validator.choices(["after signature", "before signature"])
  source_position: str = 'after signature'

  #: Allows you to override how the "view source" link is rendered into the Markdown
  #: file if a #source_linker is configured. The default is `[[view_source]]({url})`.
  source_format: str = '[[view_source]]({url})'

  #: Escape html in docstring. Default to False.
  escape_html_in_docstring: bool = False

  #: Render Novella `@anchor` tags before headings.
  render_novella_anchors: bool = False

  #: Format code rendered into Markdown code blocks with YAPF.
  format_code: bool = True

  #: The style to format code as. This can be a YAPF builtin style name or point to
  #: a file relative to the context directory (usually the working directory).
  format_code_style: str = 'pep8'

  def __post_init__(self) -> None:
    self._resolver = MarkdownReferenceResolver()

  def _is_method(self, obj: docspec.ApiObject) -> bool:
    return is_method(obj)

  def _format_arglist(self, func: docspec.Function) -> str:
    args = func.args[:]
    if self._is_method(func) and args and args[0].name == 'self':
      args.pop(0)
    return format_arglist(args)

  def _render_toc(self, fp: t.TextIO, level: int, obj: docspec.ApiObject):
    if level > self.toc_maxdepth:
      return
    object_id = self._resolver.generate_object_id(obj)
    title = self._escape(obj.name)
    if not self.add_module_prefix and isinstance(obj, docspec.Module):
      title = title.split('.')[-1]
    fp.write('  ' * level + '* [{}](#{})\n'.format(title, object_id))
    level += 1
    for child in getattr(obj, 'members', []):
      self._render_toc(fp, level, child)

  def _render_header(self, fp: t.TextIO, level: int, obj: docspec.ApiObject):
    if self.render_module_header_template and isinstance(obj, docspec.Module):
      fp.write(
        self.render_module_header_template.format(
          module_name=obj.name,
          relative_module_name=obj.name.rsplit(".", 1)[-1]
        )
      )
      return

    object_id = self._resolver.generate_object_id(obj)
    if self.use_fixed_header_levels:
      # Read the header level based on the API object type. The default levels defined
      # in the field will act as a first fallback, the level of the object inside it's
      # hierarchy is the final fallback.
      header_levels = {
        **type(self).__dataclass_fields__['header_level_by_type'].default_factory(),
        **self.header_level_by_type,
      }
      # Backwards compat for when we used "Data" instead of "Variable" which mirrors the docspec API
      header_levels['Variable'] = header_levels.get('Data', header_levels['Variable'])

      type_name = 'Method' if self._is_method(obj) else type(obj).__name__
      level = header_levels.get(type_name, level)
    if self.insert_header_anchors and not self.html_headers:
      fp.write('<a id="{}"></a>\n\n'.format(object_id))
    if self.html_headers:
      header_template = '<h{0} id="{1}">{{title}}</h{0}>'.format(level, object_id)
    else:
      header_template = level * '#' + ' {title}'
    if self.render_novella_anchors:
      fp.write(f'@anchor pydoc:' + '.'.join(x.name for x in obj.path) + '\n')
    fp.write(header_template.format(title=self._get_title(obj)))
    fp.write('\n\n')

  def _format_decorations(self, decorations: t.List[docspec.Decoration]) -> t.Iterable[str]:
    for dec in decorations:
      yield '@{}{}\n'.format(dec.name, dec.args or '')

  def _yapf_code(self, code: str) -> str:
    if not self.format_code:
      return code
    style_file = Path(self._context.directory) / self.format_code_style
    style = str(style_file) if style_file.is_file() else self.format_code_style
    return FormatCode(code, style_config=style)[0]

  def _format_function_signature(self, func: docspec.Function, override_name: str = None, add_method_bar: bool = True) -> str:
    parts: t.List[str] = []
    if self.signature_with_decorators:
      parts += self._format_decorations(func.decorations or [])
    if self.signature_python_help_style and not self._is_method(func):
      parts.append('{} = '.format(dotted_name(func)))
    parts += [x + ' ' for x in func.modifiers or []]
    if self.signature_with_def:
      parts.append('def ')
    if self.signature_class_prefix and self._is_method(func):
      parent = func.parent
      assert parent, func
      parts.append(parent.name + '.')
    parts.append((override_name or func.name))
    parts.append(format_function_signature(func, self._is_method(func)))
    result = ''.join(parts)
    result = self._yapf_code(result + ': pass').rpartition(':')[0].strip()

    if add_method_bar and self._is_method(func):
      result = '\n'.join(' | ' + l for l in result.split('\n'))
    return result

  def _format_classdef_signature(self, cls: docspec.Class) -> str:
    bases = ', '.join(map(str, cls.bases or []))
    if cls.metaclass:
      bases += ', metaclass=' + str(cls.metaclass)
    code = 'class {}({})'.format(cls.name, bases)
    if self.signature_python_help_style:
      code = dotted_name(cls) + ' = ' + code
    code = self._yapf_code(code + ': pass').rpartition(':')[0].strip()

    if cls.decorations and self.classdef_with_decorators:
      code = '\n'.join(self._format_decorations(cls.decorations)) + code
    return code

  def _format_data_signature(self, data: docspec.Variable) -> str:
    expr = str(data.value)
    code = self._yapf_code(data.name + ' = ' + expr).strip()
    if len(code) > self.data_expression_maxlength:
      code = code[:self.data_expression_maxlength] + ' ...'
    return code

  def _render_signature_block(self, fp: t.TextIO, obj: docspec.ApiObject):
    if self.classdef_code_block and isinstance(obj, docspec.Class):
      code = self._format_classdef_signature(obj)
    elif self.signature_code_block and isinstance(obj, docspec.Function):
      code = self._format_function_signature(obj, add_method_bar=self.signature_with_vertical_bar)
    elif self.data_code_block and isinstance(obj, docspec.Variable):
      code = self._format_data_signature(obj)
    else:
      return
    fp.write('```{}\n'.format('python' if self.code_lang else ''))
    fp.write(code)
    fp.write('\n```\n\n')

  def _render_object(self, fp: t.TextIO, level: int, obj: docspec.ApiObject):
    if not isinstance(obj, docspec.Module) or self.render_module_header:
      self._render_header(fp, level, obj)

    render_view_source = not isinstance(obj, (docspec.Module, docspec.Variable))

    if render_view_source:
      url = self.source_linker.get_source_url(obj) if self.source_linker else None
      source_string = self.source_format.replace('{url}', str(url)) if url else None
      if source_string and self.source_position == 'before signature':
        fp.write(source_string + '\n\n')

    self._render_signature_block(fp, obj)

    if render_view_source:
      if source_string and self.source_position == 'after signature':
        fp.write(source_string + '\n\n')

    if obj.docstring:
      docstring = html.escape(obj.docstring.content) if self.escape_html_in_docstring else obj.docstring.content
      lines = docstring.split('\n')
      if self.docstrings_as_blockquote:
        lines = ['> ' + x for x in lines]
      fp.write('\n'.join(lines))
      fp.write('\n\n')

  def _render_recursive(self, fp: t.TextIO, level: int, obj: docspec.ApiObject):
    self._render_object(fp, level, obj)
    level += 1
    for member in getattr(obj, 'members', []):
      self._render_recursive(fp, level, member)

  def _get_title(self, obj: docspec.ApiObject) -> str:
    title = obj.name
    if (self.add_method_class_prefix and self._is_method(obj)) or \
       (self.add_member_class_prefix and isinstance(obj, docspec.Variable)):
      title = (obj.parent.name + '.' + title) if obj.parent else title
    elif self.add_full_prefix and not self._is_method(obj):
      title = dotted_name(obj)
    if (not self.add_module_prefix and isinstance(obj, docspec.Module)):
      title = title.split('.')[-1]
    if isinstance(obj, docspec.Function):
      if self.signature_in_header:
        title += '(' + self._format_arglist(obj) + ')'

    if isinstance(obj, docspec.Variable) and obj.datatype and self.render_typehint_in_data_header:
      if self.code_headers:
        title += f': {obj.datatype}'
      elif self.html_headers:
        title += f': <code>{obj.datatype}</code>'
      else:
        title += f': `{obj.datatype}`'

    if self.code_headers:
      if self.html_headers or self.sub_prefix:
        if self.sub_prefix and '.' in title:
          prefix, title = title.rpartition('.')[::2]
          title = '<sub>{}.</sub>{}'.format(prefix, title)
        title = '<code>{}</code>'.format(title)
      else:
        title = '`{}`'.format(title)
    elif not self.html_headers:
      title = self._escape(title)
    if isinstance(obj, docspec.Module) and self.descriptive_module_title:
      title = 'Module ' + title
    if isinstance(obj, docspec.Class) and self.descriptive_class_title:
      if self.descriptive_class_title is True:
        title += ' Objects'
      elif self.descriptive_class_title is False:
        pass
      elif self.descriptive_class_title.startswith('$'):
        title += self.descriptive_class_title[1:]
      else:
        title = self.descriptive_class_title + title
    return title

  def _escape(self, s):
    return s.replace('_', '\\_').replace('*', '\\*')

  def render_to_string(self, modules: t.List[docspec.Module]) -> str:
    fp = io.StringIO()
    self.render_single_page(fp, modules)
    return fp.getvalue()

  def _render_to_stream(self, modules: t.List[docspec.Module], stream: t.TextIO):
    return self.render_single_page(stream, modules)

  # SinglePageRenderer

  def render_single_page(self, fp: t.TextIO, modules: t.List[docspec.Module], page_title: t.Optional[str] = None) -> None:
    if self.render_page_title:
      fp.write('# {}\n\n'.format(page_title))

    if self.render_toc:
      if self.render_toc_title:
        if self.render_page_title:
          # set to level2 since level1 is page title
          fp.write('## {}\n\n'.format(self.render_toc_title))
        else:
          fp.write('# {}\n\n'.format(self.render_toc_title))

      for m in modules:
        self._render_toc(fp, 0, m)
      fp.write('\n')
    for m in modules:
      self._render_recursive(fp, 1, m)

  # SingleObjectRenderer

  def render_object(self, fp: t.TextIO, obj: docspec.ApiObject, options: t.Dict[str, t.Any]) -> None:
    self._render_recursive(fp, 0, obj)

  # Renderer

  def get_resolver(self, modules: t.List[docspec.Module]) -> t.Optional[Resolver]:
    """
    Returns a simple #Resolver implementation. Finds cross-references in the same file.
    """

    return self._resolver

  def render(self, modules: t.List[docspec.Module]) -> None:
    if self.filename is None:
      self._render_to_stream(modules, sys.stdout)
    else:
      with io.open(self.filename, 'w', encoding=self.encoding) as fp:
        self._render_to_stream(modules, t.cast(t.TextIO, fp))

  # PluginBase

  def init(self, context: Context) -> None:
    if self.source_linker:
      self.source_linker.init(context)
    self._context = context


@dataclasses.dataclass
class MarkdownReferenceResolver(Resolver, ResolverV2):

  local: bool = True
  global_: bool = False

  def generate_object_id(self, obj: docspec.ApiObject) -> str:
    return '.'.join(o.name for o in obj.path)

  def _resolve_reference_in_members(self, obj: t.Optional[docspec.ApiObject], ref: t.List[str]) -> t.Optional[docspec.ApiObject]:
    if not obj:
      return None
    for part_name in ref:
      obj = docspec.get_member(obj, part_name)
      if not obj:
        return None
    return obj

  def _resolve_local_reference(self, scope: docspec.ApiObject, ref_split: t.List[str]) -> t.Optional[docspec.ApiObject]:
    obj: t.Optional[docspec.ApiObject] = scope
    while obj:
      resolved = self._resolve_reference_in_members(obj, ref_split)
      if resolved:
        return resolved
      obj = obj.parent
    return None

  # Resolver

  def resolve_ref(self, scope: docspec.ApiObject, ref: str) -> t.Optional[str]:
    target = self._resolve_local_reference(scope, ref.split('.'))
    if target:
      return '#' + self.generate_object_id(target)
    return None

  # ResolverV2

  def resolve_reference(self, suite: ApiSuite, scope: docspec.ApiObject, ref: str) -> t.Optional[docspec.ApiObject]:
    """ Resolves the reference by searching in the members of *scope* or any of its parents. """

    # TODO (@NiklasRosenstein): Support resolving indirections

    ref_split = ref.split('.')

    resolved = self._resolve_local_reference(scope, ref_split)
    if resolved:
      return resolved

    if self.global_:
      def _recurse(obj: docspec.ApiObject) -> t.Optional[docspec.ApiObject]:
        resolved = self._resolve_reference_in_members(obj, ref_split)
        if resolved:
          return resolved
        if isinstance(obj, docspec.HasMembers):
          for member in obj.members:
            resolved = _recurse(member)
            if resolved:
              return resolved
        return None

      for module in suite:
        resolved = _recurse(module)
        if resolved:
          return resolved

    return None




# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import dataclasses
import typing as t

import docspec

from pydoc_markdown.interfaces import Processor, Resolver


@dataclasses.dataclass
class FilterProcessor(Processor):
  """
  The `filter` processor removes module and class members based on certain criteria.

  Example configuration:

  ```yaml
  - type: filter
    expression: not name.startswith('_') and default()
    documented_only: false
  ```

  ### Options
  """

  #: A Python expression that is evaluated given the variables `name`, `obj` and `default`
  #: and is expected to return a boolean to indicate whether the #docspec.ApiObject should
  #: be kept or removed. If specified, the expression is the ultimate truth for determining
  #: the keep-or-remove state of a node. Using `'default()'` as the expression has the same
  #: semantic as not specifying this field. Default: `null`
  expression: t.Optional[str] = None

  #: Keep only API objects that have docstrings. Default: `true`
  documented_only: bool = True

  #: Exclude API objects that appear to be private members (i.e. their name begins with
  #: and underscore but does not end with one). Default: `true`
  exclude_private: bool = True

  #: Exclude special members (e.g.` __path__`, `__annotations__`, `__name__` and `__all__`).
  #: Default: `true`
  exclude_special: bool = True

  #: Do not filter #docspec.Module objects. Default: `true`
  do_not_filter_modules: bool = True

  #: Skip modules with no content. Default: `false`.
  skip_empty_modules: bool = False

  SPECIAL_MEMBERS = ('__path__', '__annotations__', '__name__', '__all__')

  def process(self, modules: t.List[docspec.Module], resolver: t.Optional[Resolver]) -> None:
    def m(obj):
      return self._match(obj)
    docspec.filter_visit(t.cast(t.List[docspec.ApiObject], modules), m, order='post')

  def _match(self, obj: docspec.ApiObject) -> bool:
    members = getattr(obj, 'members', [])

    def _check():
      if members:
        return True
      if self.skip_empty_modules and isinstance(obj, docspec.Module) and not members:
        return False
      if self.do_not_filter_modules and isinstance(obj, docspec.Module):
        return True
      if self.documented_only and not obj.docstring:
        return False
      if self.exclude_private and obj.name.startswith('_') and not obj.name.endswith('_'):
        return False
      if self.exclude_special and obj.name in self.SPECIAL_MEMBERS:
        return False
      return True

    if self.expression:
      scope = {'name': obj.name, 'obj': obj, 'default': _check}
      return bool(eval(self.expression, scope))  # pylint: disable=eval-used

    return _check()




from __future__ import annotations

import logging
import re
import typing as t
from pathlib import Path

from nr.util.functional import assure
from nr.util.plugins import load_entrypoint
from novella.markdown.preprocessor import MarkdownFile, MarkdownFiles, MarkdownPreprocessor
from novella.markdown.tagparser import parse_block_tags, parse_inline_tags, replace_tags, Tag
from novella.repository import RepositoryType, detect_repository
from pydoc_markdown.contrib.source_linkers import git as git_source_linkers
from pydoc_markdown.contrib.processors.crossref import CrossrefProcessor
from pydoc_markdown.contrib.processors.filter import FilterProcessor
from pydoc_markdown.contrib.processors.smart import SmartProcessor
from pydoc_markdown.interfaces import Context, Loader, Processor, SingleObjectRenderer, SourceLinker
from pydoc_markdown.contrib.loaders.python import PythonLoader
from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer, MarkdownReferenceResolver
from pydoc_markdown.util.docspec import ApiSuite

logger = logging.getLogger(__name__)


def autodetect_source_linker() -> t.Optional[SourceLinker]:
  repo = detect_repository(Path.cwd())
  if not repo or repo.type != RepositoryType.GIT:
    return None

  if 'github.com' in repo.url:
    name = assure(re.search(r'github.com/([^/]+/[^/]+)', repo.url)).group(1)
    return git_source_linkers.GithubSourceLinker(root=repo.root, repo=name, use_branch=True)
  elif 'gitlab.com' in repo.url:
    name = assure(re.search(r'gitlab.com/(.*)', repo.url)).group(1)
    return git_source_linkers.GitlabSourceLinker(root=repo.root, repo=name, use_branch=True)
  elif 'gitea.com' in repo.url:
    name = assure(re.search(r'gitea.com/([^/]+/[^/]+)', repo.url)).group(1)
    return git_source_linkers.GiteaSourceLinker(root=repo.root, repo=name, use_branch=True)

  # TODO (@NiklasRosenstein): BitBucket
  return None


class PydocTagPreprocessor(MarkdownPreprocessor):
  """ Implements the `@pydoc` and `@pylink` tag when using Novella for Markdown preprocessing

  This preprocessor precedes the Novella built-in "anchor" preprocessor as it generates `@anchor` and `{@link}` tags.
  """

  _loader: Loader
  _processors: t.List[Processor]
  _renderer: SingleObjectRenderer
  _suite: ApiSuite | None = None

  def __post_init__(self) -> None:
    # Heuristic to provide a sensible default configuration of the plugin.
    if Path.cwd().name.lower() in ('docs', 'documentation'):
      search_path = ['../src', '..']
    else:
      search_path = ['src', '.']

    source_linker = autodetect_source_linker()
    self._loader = PythonLoader(search_path=search_path)
    self._processors = [
      FilterProcessor(),
      SmartProcessor(),
      # We return the entire link formatted as a Novella {@link} tag in #resolve_ref().
      CrossrefProcessor(resolver_v2=MarkdownReferenceResolver(global_=True)),
    ]
    self._renderer = MarkdownRenderer(
      source_linker=source_linker,
      render_novella_anchors=True,
      render_module_header=False,
      descriptive_class_title='Class ',
    )

  @t.overload
  def loader(self) -> Loader: ...

  @t.overload
  def loader(self, loader: str | Loader, closure: t.Callable[[Loader], t.Any] | None) -> None: ...

  def loader(self, loader=None, closure=None):
    if loader is not None:
      if isinstance(loader, str):
        loader = t.cast(Loader, load_entrypoint('pydoc_markdown.interfaces.Loader', loader)())
      if closure is not None:
        closure(loader)
      self._loader = loader
    else:
      if self._loader is None:
        raise RuntimeError('no loader has been configured yet')
      return self._loader

  @t.overload
  def renderer(self) -> SingleObjectRenderer: ...

  @t.overload
  def renderer(self, renderer: str | Loader, closure: t.Callable[[Loader], t.Any] | None) -> None: ...

  def renderer(self, renderer=None, closure=None):
    if renderer is not None:
      if isinstance(renderer, str):
        renderer = load_entrypoint('pydoc_markdown.interfaces.Renderer', renderer)()
        if not isinstance(renderer, SingleObjectRenderer):
          raise RuntimeError(f'not a SingleObjectRenderer: {type(renderer).__name__}')
      if closure is not None:
        closure(renderer)
      self._renderer = renderer
    else:
      return self._renderer

  # MarkdownPreprocessor

  def setup(self) -> None:
    if self.dependencies is None and self.predecessors is None:
      self.precedes('anchor')

  def process_files(self, files: MarkdownFiles) -> None:
    context = Context(str(Path.cwd()))
    self._loader.init(context)
    self._renderer.init(context)

    if self._suite is None:
      modules = list(self._loader.load())
      for processor in self._processors:
        processor.process(modules, self)
      self._suite = ApiSuite(modules)

    for file in files:
      tags = [t for t in parse_block_tags(file.content) if t.name == 'pydoc']
      file.content = replace_tags(file.content, tags, lambda t: self._replace_pydoc_tag(self._suite, file, t))
      tags = [t for t in parse_inline_tags(file.content) if t.name == 'pylink']
      file.content = replace_tags(file.content, tags, lambda t: self._replace_pylink_tag(t))

  def _replace_pydoc_tag(self, suite: ApiSuite, file: MarkdownFile, tag: Tag) -> str | None:
    fqn = tag.args.strip()
    objects = suite.resolve_fqn(fqn)
    if len(objects) > 1:
      logger.warning('  found multiple matches for Python FQN <fg=cyan>%s</fg>', fqn)
    elif not objects:
      logger.warning('  found no match for Python FQN <fg=cyan>%s</fg>', fqn)
      return None

    import io
    fp = io.StringIO()
    self._renderer.render_object(fp, objects[0], tag.options)
    return self.action.repeat(file.path, file.output_path, fp.getvalue())

  def _replace_pylink_tag(self, tag: Tag) -> str | None:
    return f'{{@link pydoc:{tag.args.strip()}}}'



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
This module defines the interfaces that can to be implemented for
Pydoc-Markdown to implement custom loaders for documentation data,
processors or renderers.
"""

import abc
import subprocess
import typing as t

import docspec
from databind.core import annotations as A

if t.TYPE_CHECKING:
  from pydoc_markdown.util.docspec import ApiSuite


class Context:
  """
  Context data that is passed to plugins when they are loaded.
  """

  def __init__(self, directory: str) -> None:
    self.directory = directory


class PluginBase(abc.ABC):

  def init(self, context: Context) -> None: pass


@A.unionclass(A.unionclass.Subtypes.chain(
  A.unionclass.Subtypes.entrypoint('pydoc_markdown.interfaces.Loader'),
  A.unionclass.Subtypes.import_()), style=A.unionclass.Style.flat)
class Loader(PluginBase):
  """
  This interface describes an object that is capable of loading documentation
  data. The location from which the documentation is loaded must be defined
  with the configuration class.
  """

  @abc.abstractmethod
  def load(self) -> t.Iterable[docspec.Module]: ...


class LoaderError(Exception):
  pass


class Resolver(abc.ABC):
  """
  A resolver can be used by a #Processor to replace cross references with a hyperlink.
  """

  @abc.abstractmethod
  def resolve_ref(self, scope: docspec.ApiObject, ref: str) -> t.Optional[str]: ...


class ResolverV2(abc.ABC):
  """ New style interface for resolving based on a text ref from in the context of a #docspec.ApiObject
  to find another. This is different from #Resolver because it returns the resolved object directly, instead
  of some string representation of it.
  """

  @abc.abstractmethod
  def resolve_reference(self, suite: 'ApiSuite', scope: docspec.ApiObject, ref: str) -> t.Optional[docspec.ApiObject]: ...


@A.unionclass(A.unionclass.Subtypes.chain(
  A.unionclass.Subtypes.entrypoint('pydoc_markdown.interfaces.Processor'),
  A.unionclass.Subtypes.import_()), style=A.unionclass.Style.flat)
class Processor(PluginBase):
  """
  A processor is an object that takes a list of #docspec.Module#s as an input and
  transforms it in an arbitrary way. This usually processes docstrings to convert from
  various documentation syntaxes to plain Markdown.
  """

  @abc.abstractmethod
  def process(self, modules: t.List[docspec.Module], resolver: t.Optional[Resolver]) -> None: ...


@A.unionclass(A.unionclass.Subtypes.chain(
  A.unionclass.Subtypes.entrypoint('pydoc_markdown.interfaces.Renderer'),
  A.unionclass.Subtypes.import_()), style=A.unionclass.Style.flat)
class Renderer(PluginBase):
  """
  A renderer is an object that takes a list of #docspec.Module#s as an input and produces
  output files or writes to stdout. It may also expose additional command-line arguments.
  There can only be one renderer at the end of the processor chain.

  Note that sometimes a renderer may need to perform some processing before the render step.
  To keep the possibility open that a renderer may implement generic processing that could
  used without the actual renderering functionality, #Renderer is a subclass of #Processor.
  """

  def process(self, modules: t.List[docspec.Module], resolver: t.Optional[Resolver]) -> None: pass

  def get_resolver(self, modules: t.List[docspec.Module]) -> t.Optional[Resolver]: return None

  @abc.abstractmethod
  def render(self, modules: t.List[docspec.Module]) -> None: ...


class SinglePageRenderer(PluginBase):
  """
  Interface for rendering a single page.
  """

  @abc.abstractmethod
  def render_single_page(self, fp: t.TextIO, modules: t.List[docspec.Module],
                         page_title: t.Optional[str] = None) -> None: ...


class SingleObjectRenderer(PluginBase):
  """
  Interface for rendering a single #docspec.ApiObject.
  """

  @abc.abstractmethod
  def render_object(self, fp: t.TextIO, obj: docspec.ApiObject, options: t.Dict[str, t.Any]) -> None: ...


class Server(abc.ABC):
  """
  This interface describes an object that can start a server process for
  live-viewing generated documentation in the browser. #Renderer
  implementations may additionally implement this interface to advocate their
  compatibility with the `--server` and `--open` options of the pydoc-markdown
  CLI.
  """

  @abc.abstractmethod
  def get_server_url(self) -> str: ...

  @abc.abstractmethod
  def start_server(self) -> subprocess.Popen: ...

  def reload_server(self, process: subprocess.Popen) -> subprocess.Popen:
    """
    Called when the files generated by pydoc-markdown have been updated.
    This gives the implementation a chance to reload the server process.
    The default implementation returns the *process* unchanged. Returning
    #None will automatically call #start_server() afterwards.
    """

    return process


class Builder(abc.ABC):
  """
  This interface can be implemented additionally to the #Renderer interface to
  indicate that the renderer supports building another produce after the markdown
  files have been rendered.
  """

  @abc.abstractmethod
  def build(self, site_dir: str) -> None:
    """
    Invoke the build. The *site_dir* is the directory in which the output files should be
    placed. Otherwise, the directory may be determined by the builder.
    """


@A.unionclass(A.unionclass.Subtypes.chain(
  A.unionclass.Subtypes.entrypoint('pydoc_markdown.interfaces.SourceLinker'),
  A.unionclass.Subtypes.import_()), style=A.unionclass.Style.flat)
class SourceLinker(PluginBase):
  """
  This interface is used to determine the URL to the source of an API object. Renderers
  can use it to place a link to the source in the generated documentation.
  """

  @abc.abstractmethod
  def get_source_url(self, obj: docspec.ApiObject) -> t.Optional[str]: ...




"""
__Important__: The Jinja2 renderer API and configuration is not stable. Your config/templates may break
at any point on future Pydoc-Markdown versions.
"""

import dataclasses as D
import fnmatch
import logging
import os
import typing as t

import docspec
from docspec_python import format_arglist
import jinja2

from pydoc_markdown.contrib.renderers.markdown import MarkdownReferenceResolver
from pydoc_markdown.interfaces import Renderer, Resolver
from pydoc_markdown.util.docspec import format_function_signature, get_members_of_type, get_object_description

T = t.TypeVar('T')
log = logging.getLogger(__name__)


class Args(t.Dict[str, t.Any]):

  def get_render_args(self, modules: t.List[docspec.Module]) -> t.Dict[str, t.Any]:
    args = dict(self)
    if 'module' in args:
      try:
        args['module'] = next((m for m in modules if m.name == args['module']))
      except StopIteration:
        raise ValueError(f'module {args["module"]} not found')
    if 'modules' in args:
      args['modules'] = [m for m in modules if m.name in args['modules'] or any(fnmatch.fnmatch(m.name, x) for x in args['modules'])]
    return args


@D.dataclass
class RenderBlock:

  #: The path to the Jinja2 template that is used to render the output files.
  template: str

  #: A mapping for filename (without suffix) to a list of Module selectors.
  produces: t.Dict[str, Args]

  #: Settings for the Jinja2 Environment.
  jinja2_environment_settings: t.Dict[str, t.Any] = D.field(default_factory=dict)


@D.dataclass
class Jinja2Renderer(Renderer):

  #: Render instructions.
  renders: t.List[RenderBlock]

  #: Build directory where all the files are produced.
  build_directory: str = 'build/docs'

  def render(self, modules: t.List[docspec.Module]) -> None:
    # TODO (@NiklasRosenstein): Clean render support

    resolver = MarkdownReferenceResolver()
    os.makedirs(self.build_directory, exist_ok=True)

    for render in self.renders:
      env = jinja2.Environment(loader=jinja2.FileSystemLoader('.'), **render.jinja2_environment_settings)
      setup_env(env)
      env.filters['uid'] = resolver.generate_object_id
      template = env.get_template(render.template)
      for filename, args in render.produces.items():
        filename = os.path.join(self.build_directory, filename + '.md')
        log.info('Writing %s', filename)
        with open(filename, 'w') as fp:
          fp.write(template.render(**Args(args).get_render_args(modules)))  # TODO

  def get_resolver(self, modules: t.List[docspec.Module]) -> t.Optional[Resolver]:
    return MarkdownReferenceResolver()


def setup_env(env: jinja2.Environment) -> None:
  env.filters['classes'] = lambda modules: get_members_of_type(modules, docspec.Class)
  env.filters['functions'] = lambda objs: get_members_of_type(objs, docspec.Function)
  env.filters['attrs'] = lambda objs: get_members_of_type(objs, docspec.Variable)
  env.filters['indent'] = _indent_filter
  env.filters['blockquote'] = _blockquote_filter
  env.filters['first_line'] = _first_line_filter
  env.filters['format_arglist'] = format_arglist
  env.filters['format_function_signature'] = format_function_signature
  env.filters['describe'] = lambda obj: get_object_description(obj)


def _indent_filter(text: t.Optional[str], level: int = 1, ) -> str:
  if not text: return ''
  lines = text.splitlines()
  return '\n'.join(lines[:1] + list('    ' * level + l for l in lines[1:]))


def _blockquote_filter(text: t.Optional[str]) -> str:
  if not text: return ''
  return '> ' + '\n> '.join(text.splitlines())


def _first_line_filter(text: t.Optional[str]) -> str:
  if not text: return ''
  return text.splitlines()[0]



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import dataclasses
import logging
import typing as t

import docspec
import docstring_parser

from pydoc_markdown.interfaces import Processor, Resolver

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class _ParamLine:
  """
  Helper data class for holding details of Sphinx arguments.
  """

  name: str
  docs: str
  type: t.Optional[str] = None


def generate_sections_markdown(lines, sections):
  for key, section in sections.items():
    if section:
      if lines and lines[-1]:
        lines.append('')
      lines.extend(['**{}**:'.format(key), ''])
      lines.extend(section)



@dataclasses.dataclass
class SphinxProcessor(Processor):
  """
  This processor parses ReST/Sphinx-style function documentation and converts it into
  Markdown syntax.

  Example:

  ```
  :param arg1: This is the first argument.
  :raise ValueError: If *arg1* is a bad value.
  :return: An `int` that represents an interesting value.
  ```

  Renders as:

  :param arg1: This is the first argument.
  :raise ValueError: If *arg1* is a bad value.
  :return: An `int` that represents an interesting value.

  @doc:fmt:sphinx
  """

  _KEYWORDS = {
    'Arguments': [
      'arg',
      'argument',
      'param',
      'parameter',
      'type',
    ],
    'Returns': [
      'return',
      'returns',
      'rtype',
    ],
    'Raises': [
      'raises',
      'raise',
    ]
  }

  def check_docstring_format(self, docstring: str) -> bool:
    return any(f':{k}' in docstring for _, value in self._KEYWORDS.items() for k in value)

  def process(self, modules: t.List[docspec.Module], resolver: t.Optional[Resolver]) -> None:
    docspec.visit(modules, self._process)

  def _convert_raises(self, raises: t.List[docstring_parser.common.DocstringRaises]) -> list:
    """Convert a list of DocstringRaises from docstring_parser to markdown lines

    :return: A list of markdown formatted lines
    """
    converted_lines = []
    for entry in raises:
      converted_lines.append('- `{}`: {}'.format(entry.type_name, entry.description))
    return converted_lines

  def _convert_params(self, params: t.List[docstring_parser.common.DocstringParam]) -> list:
    """Convert a list of DocstringParam to markdown lines.

    :return: A list of markdown formatted lines
    """
    converted = []
    for param in params:
      if param.type_name is None:
        converted.append(
          '- `{name}`: {description}'.format(name=param.arg_name, description=param.description)
        )
      else:
        converted.append(
          '- `{name}` (`{type}`): {description}'.format(name=param.arg_name,
                                                        type=param.type_name,
                                                        description=param.description)
        )
    return converted

  def _convert_returns(self, returns: t.Optional[docstring_parser.common.DocstringReturns]) -> str:
    """Convert a DocstringReturns object to a markdown string.

    :return: A markdown formatted string
    """
    if returns is not None:
      if returns.type_name:
        type_data = '`{}`: '.format(returns.type_name)
      else:
        type_data = ''
      return_data = type_data + (returns.description or '')
    else:
      return_data = ''
    return return_data

  def _process(self, node: docspec.ApiObject) -> None:
    if not node.docstring:
      return

    lines = []
    components: t.Dict[str, t.List[str]] = {}

    parsed_docstring = docstring_parser.parse(node.docstring.content, docstring_parser.DocstringStyle.REST)
    components['Arguments'] = self._convert_params(parsed_docstring.params)
    components['Raises'] = self._convert_raises(parsed_docstring.raises)
    return_doc = self._convert_returns(parsed_docstring.returns)
    if return_doc:
      components['Returns'] = [return_doc]

    if parsed_docstring.short_description:
      lines.append(parsed_docstring.short_description)
      lines.append('')
    if parsed_docstring.long_description:
      lines.append(parsed_docstring.long_description)
      lines.append('')

    generate_sections_markdown(lines, components)
    node.docstring.content = '\n'.join(lines)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import os
import typing as t
from pathlib import Path

import databind.core.dataclasses as dataclasses
import docspec
from nr.util.git import Git, NoCurrentBranchError

from pydoc_markdown.interfaces import Context, SourceLinker

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class BaseGitSourceLinker(SourceLinker):
  """
  This is the base source linker for code in a Git repository.
  """

  #: The root directory, relative to the context directory. The context directory
  #: is usually the directory that contains Pydoc-Markdown configuration file.
  #: If not set, the project root will be determined from the Git repository in
  #: the context directory.
  root: t.Optional[str] = None

  #: Use the branch name instead of the current SHA to generate URLs.
  use_branch: bool = False

  def __post_init__(self) -> None:
    self._project_root: t.Optional[str] = None
    self._branch: t.Optional[str] = None
    self._sha: t.Optional[str] = None

  def get_context_vars(self) -> t.Dict[str, str]:
    return {}

  def get_url_template(self) -> str:
    raise NotImplementedError

  # SourceLinker

  def get_source_url(self, obj: docspec.ApiObject) -> t.Optional[str]:
    """
    Compute the URL in the GitHub/Gitea #repo for the API object *obj*.
    """

    assert self._project_root is not None

    if not obj.location or not obj.location.filename:
      return None

    # Compute the path relative to the project root.
    try:
      rel_path = Path(obj.location.filename).relative_to(self._project_root)
    except ValueError:
      logger.debug('Ignored API object %s, path points outside of project root.', obj.name)
      return None

    context_vars = self.get_context_vars()
    context_vars['path'] = str(rel_path)
    context_vars['sha'] = (self._branch if self.use_branch else self._sha) or '?'
    context_vars['lineno'] = str(obj.location.lineno)

    url = self.get_url_template().format(**context_vars)

    logger.debug('Calculated URL for API object %s is %s', obj.name, url)
    return url

  # PluginBase

  def init(self, context: Context) -> None:
    """
    This is called before the rendering process. The source linker computes the project
    root and Git SHA at this stage before #get_source_url() is called.
    """

    git = Git(context.directory)

    if self.root:
      self._project_root = os.path.join(context.directory, self.root)
    else:
      project_root = git.get_toplevel()
      if not project_root:
        raise RuntimeError(f'Path "%s" is not in a Git repository', context.directory)
      self._project_root = project_root

    self._sha = git.rev_parse('HEAD')
    self._branch = None

    if self.use_branch:
      try:
        self._branch = git.get_current_branch_name()
        if '(' in self._branch:
          raise NoCurrentBranchError  # For older versions of nr.util which returned the detached HEAD branch
      except NoCurrentBranchError:
        logger.warning('Repository is not currently on a branch, falling back to SHA')
        self.use_branch = False

    logger.debug('project_root = %r', self._project_root)
    logger.debug('sha = %r', self._sha)
    logger.debug('branch = %r', self._branch)


@dataclasses.dataclass
class BaseGitServiceSourceLinker(BaseGitSourceLinker):

  #: The repository name, formatted as `owner/repo`.
  repo: str

  #: The host name of the Git service.
  host: str

  def get_context_vars(self) -> t.Dict[str, str]:
    return {'repo': self.repo, 'host': self.host}


@dataclasses.dataclass
class GitSourceLinker(BaseGitSourceLinker):
  """
  This source linker allows you to define your own URL template to generate a permalink for
  an API object.
  """

  #: The URL template as a Python format string. Available variables are "path", "sha" and "lineno".
  url_template: str

  def get_url_template(self) -> str:
    return self.url_template


@dataclasses.dataclass
class GiteaSourceLinker(BaseGitServiceSourceLinker):
  """
  Source linker for Git repositories hosted on gitea.com or any self-hosted Gitea instance.
  """

  host: str = "gitea.com"

  def get_url_template(self) -> str:
    return 'https://{host}/{repo}/src/commit/{sha}/{path}#L{lineno}'


@dataclasses.dataclass
class GithubSourceLinker(BaseGitServiceSourceLinker):
  """
  Source linker for Git repositories hosted on github.com or any self-hosted GitHub instance.
  """

  host: str = "github.com"

  def get_url_template(self) -> str:
    return 'https://{host}/{repo}/blob/{sha}/{path}#L{lineno}'


@dataclasses.dataclass
class GitlabSourceLinker(BaseGitServiceSourceLinker):
  """
  Source linker for Git repositories hosted on gitlab.com or any self-hosted Gitlab instance.
  """

  host: str = "gitlab.com"

  def get_url_template(self) -> str:
    return 'https://{host}/{repo}/-/blob/{sha}/{path}#L{lineno}'


@dataclasses.dataclass
class BitbucketSourceLinker(BaseGitServiceSourceLinker):
  """
  Source linker for Git repositories hosted on bitbucket.org or any self-hosted Bitbucket instance.
  """

  host: str = "bitbucket.org"

  def get_url_template(self) -> str:
    return 'https://{host}/{repo}/src/{sha}/{path}#lines-{lineno}'




from __future__ import annotations

import typing as t
import typing_extensions as te

import docspec
import docspec_python

from nr.util.generic import T
from nr.util import Stream


def get_members_of_type(objs: t.Union[docspec.ApiObject, t.List[docspec.ApiObject]], type_: t.Type[T]) -> t.List[T]:
  if isinstance(objs, docspec.HasMembers):
    return [x for x in objs.members if isinstance(x, type_)]
  elif isinstance(objs, docspec.ApiObject):
    return []
  else:
    return Stream(get_members_of_type(x, type_) for x in objs).concat().collect()


def format_function_signature(func: docspec.Function, exclude_self: bool = False) -> str:
  assert isinstance(func, docspec.Function), type(func)
  args = func.args[:]
  if exclude_self and args and args[0].name == 'self':
    args.pop(0)
  sig = f'({docspec_python.format_arglist(args)})'
  if func.return_type:
    sig += f' -> {func.return_type}'
  return sig


def is_function(obj: docspec.ApiObject) -> te.TypeGuard[docspec.Function]:
  return isinstance(obj, docspec.Function)


def is_method(obj: docspec.ApiObject) -> te.TypeGuard[docspec.Function]:
  return is_function(obj) and isinstance(obj.parent, docspec.Class)


def is_property(obj: docspec.ApiObject) -> te.TypeGuard[docspec.Function]:
  return is_function(obj) and any(d.name == 'property' for d in obj.decorations or [])


def is_attr(obj: docspec.ApiObject) -> te.TypeGuard[docspec.Variable]:
  return isinstance(obj, docspec.Variable) and isinstance(obj.parent, docspec.Class)


def get_object_description(obj: docspec.ApiObject) -> str:
  """
  Determines the best matching description of *obj*. Returns a string such as "module", "class", "abstract class",
  "enum", "attribute", "property", "function", "method", "abstract method", "class method", "static method" and
  "abstract class method".
  """

  if isinstance(obj, docspec.Module):
    return 'module'
  elif isinstance(obj, docspec.Class):
    if any(x in (obj.bases or ()) for x in ('abc.ABC', 'ABC')) or \
       any(x in (obj.metaclass or ()) for x in ('abc.ABCMeta', 'ABCMeta')):
      return 'abstract class'
    return 'class'
  elif isinstance(obj, docspec.Function):
    if any(x.name == 'property' for x in obj.decorations or []):
      return 'property'
    if any(x.name in ('abc.abstractproperty', 'abstractproperty') for x in obj.decorations or []):
      return 'abstract property'
    if any(x.name in ('abc.abstractmethod', 'abstractmethod') for x in obj.decorations or []):
      return 'abstract method'
    if any(x.name in ('abc.abstractclassmethod', 'abstractclassmethod') for x in obj.decorations or []):
      return 'abstract class method'
    if any(x.name == 'classmethod' for x in obj.decorations or []):
      return 'class method'
    if any(x.name == 'staticmethod' for x in obj.decorations or []):
      return 'static method'
    if is_method(obj):
      return 'method'
    return 'function'
  elif isinstance(obj, docspec.Variable):
    if is_attr(obj):
      return 'attribute'
    return 'data'
  else:
    assert False, type(obj)


class ApiSuite:
  """ Container for all loaded API objects. """

  def __init__(self, modules: t.List[docspec.Module]) -> None:
    self._modules = modules

  def resolve_fqn(self, fqn: str) -> t.List[docspec.ApiObject]:

    def _match(results: list[docspec.ApiObject]) -> t.Callable[[docspec.ApiObject], t.Any]:
      def matcher(obj: docspec.ApiObject) -> None:
        current_fqn = '.'.join(y.name for y in obj.path)
        if current_fqn == fqn:
          results.append(obj)
      return matcher

    results: list[docspec.ApiObject] = []
    docspec.visit(self._modules, _match(results))
    return results

  def __iter__(self) -> t.Iterator[docspec.Module]:
    return iter(self._modules)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Loads Python source code.
"""

import dataclasses
import logging
import os
import sys
import typing as t

import docspec
import docspec_python

from pydoc_markdown.interfaces import Context, Loader

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class PythonLoader(Loader):
  """
  This implementation of the #Loader interface parses Python modules and packages using
  #docspec_python. See the options below to control which modules and packages are being
  loaded and how to configure the parser.

  With no #modules or #packages set, the #PythonLoader will discover available modules
  in the current and `src/` directory.

  __lib2to3 Quirks__

  Pydoc-Markdown doesn't execute your Python code but instead relies on the
  `lib2to3` parser. This means it also inherits any quirks of `lib2to3`.

  _List of known quirks_

  * A function argument in Python 3 cannot be called `print` even though
    it is legal syntax
  """

  #: A list of module names that this loader will search for and then parse.
  #: The modules are searched using the #sys.path of the current Python
  # interpreter, unless the #search_path option is specified.
  modules: t.Optional[t.List[str]] = None

  #: A list of package names that this loader will search for and then parse,
  #: including all sub-packages and modules.
  packages: t.Optional[t.List[str]] = None

  #: The module search path. If not specified, the current #sys.path is
  #: used instead. If any of the elements contain a `*` (star) symbol, it
  #: will be expanded with #sys.path.
  search_path: t.Optional[t.List[str]] = None

  #: List of modules to ignore when using module discovery on the #search_path.
  ignore_when_discovered: t.List[str] = dataclasses.field(default_factory=lambda: ['test', 'tests', 'setup'])

  #: Options for the Python parser.
  parser: docspec_python.ParserOptions = dataclasses.field(default_factory=docspec_python.ParserOptions)

  #: The encoding to use when reading the Python source files.
  encoding: t.Optional[str] = None

  def __post_init__(self) -> None:
    self._context: t.Optional[Context] = None

  def get_effective_search_path(self) -> t.List[str]:
    if self.search_path is None:
      search_path = ['.', 'src'] if self.modules is None else list(sys.path)
    else:
      search_path = list(self.search_path)
      if '*' in search_path:
        index = search_path.index('*')
        search_path[index:index+1] = sys.path
    assert self._context is not None
    return [os.path.join(self._context.directory, x) for x in search_path]

  # Loader

  def load(self) -> t.Iterable[docspec.Module]:
    search_path = self.get_effective_search_path()
    modules = list(self.modules or [])
    packages = list(self.packages or [])
    do_discover = (self.modules is None and self.packages is None)

    if do_discover:
      for path in search_path:
        try:
          discovered_items = list(docspec_python.discover(path))
        except FileNotFoundError:
          continue

        logger.info(
          'Discovered Python modules %s and packages %s in %r.',
          [x.name for x in discovered_items if isinstance(x, docspec_python.DiscoveryResult.Module)],
          [x.name for x in discovered_items if isinstance(x, docspec_python.DiscoveryResult.Package)],
          path,
        )

        for item in discovered_items:
          if item.name in self.ignore_when_discovered:
            continue
          if isinstance(item, docspec_python.DiscoveryResult.Module):
            modules.append(item.name)
          elif isinstance(item, docspec_python.DiscoveryResult.Package):
            packages.append(item.name)

    logger.info(
      'Load Python modules (search_path: %r, modules: %r, packages: %r, discover: %s)',
      search_path, modules, packages, do_discover
    )

    return docspec_python.load_python_modules(
      modules=modules,
      packages=packages,
      search_path=search_path,
      options=self.parser,
      encoding=self.encoding,
    )

  # PluginBase

  def init(self, context: Context) -> None:
    self._context = context



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pydoc_markdown',
 'pydoc_markdown.contrib',
 'pydoc_markdown.contrib.loaders',
 'pydoc_markdown.contrib.processors',
 'pydoc_markdown.contrib.renderers',
 'pydoc_markdown.contrib.source_linkers',
 'pydoc_markdown.novella',
 'pydoc_markdown.util']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3,<6.0',
 'click>=7.1,<9.0',
 'databind>=1.5.0,<2.0.0',
 'docspec-python>=2.0.0a1,<3.0.0',
 'docspec>=2.0.0a1,<3.0.0',
 'docstring-parser>=0.11,<0.12',
 'jinja2>=3.0.0,<4.0.0',
 'nr.util>=0.7.5,<1.0.0',
 'requests>=2.23.0,<3.0.0',
 'tomli>=2.0.0,<3.0.0',
 'tomli_w>=1.0.0,<2.0.0',
 'watchdog',
 'yapf>=0.30.0']

entry_points = \
{'console_scripts': ['pydoc-markdown = pydoc_markdown.main:cli'],
 'novella.markdown.preprocessors': ['pydoc = '
                                    'pydoc_markdown.novella.preprocessor:PydocTagPreprocessor'],
 'pydoc_markdown.interfaces.Loader': ['python = '
                                      'pydoc_markdown.contrib.loaders.python:PythonLoader'],
 'pydoc_markdown.interfaces.Processor': ['crossref = '
                                         'pydoc_markdown.contrib.processors.crossref:CrossrefProcessor',
                                         'filter = '
                                         'pydoc_markdown.contrib.processors.filter:FilterProcessor',
                                         'google = '
                                         'pydoc_markdown.contrib.processors.google:GoogleProcessor',
                                         'pydocmd = '
                                         'pydoc_markdown.contrib.processors.pydocmd:PydocmdProcessor',
                                         'smart = '
                                         'pydoc_markdown.contrib.processors.smart:SmartProcessor',
                                         'sphinx = '
                                         'pydoc_markdown.contrib.processors.sphinx:SphinxProcessor'],
 'pydoc_markdown.interfaces.Renderer': ['docusaurus = '
                                        'pydoc_markdown.contrib.renderers.docusaurus:DocusaurusRenderer',
                                        'hugo = '
                                        'pydoc_markdown.contrib.renderers.hugo:HugoRenderer',
                                        'jinja2 = '
                                        'pydoc_markdown.contrib.renderers.jinja2:Jinja2Renderer',
                                        'markdown = '
                                        'pydoc_markdown.contrib.renderers.markdown:MarkdownRenderer',
                                        'mkdocs = '
                                        'pydoc_markdown.contrib.renderers.mkdocs:MkdocsRenderer'],
 'pydoc_markdown.interfaces.SourceLinker': ['bitbucket = '
                                            'pydoc_markdown.contrib.source_linkers.git:BitbucketSourceLinker',
                                            'git = '
                                            'pydoc_markdown.contrib.source_linkers.git:GitSourceLinker',
                                            'gitea = '
                                            'pydoc_markdown.contrib.source_linkers.git:GiteaSourceLinker',
                                            'github = '
                                            'pydoc_markdown.contrib.source_linkers.git:GithubSourceLinker',
                                            'gitlab = '
                                            'pydoc_markdown.contrib.source_linkers.git:GitlabSourceLinker']}

setup_kwargs = {
    'name': 'pydoc-markdown',
    'version': '4.6.3',
    'description': 'Create Python API documentation in Markdown format.',
    'long_description': '# Pydoc-Markdown\n\n![Python versions](https://img.shields.io/pypi/pyversions/pydoc-markdown?style=for-the-badge)\n[![Pypi version](https://img.shields.io/pypi/v/pydoc-markdown?style=for-the-badge)](https://pypi.org/project/pydoc-markdown/)\n[![Build status](https://img.shields.io/github/workflow/status/NiklasRosenstein/pydoc-markdown/Python%20package?style=for-the-badge)](https://github.com/NiklasRosenstein/pydoc-markdown/actions)\n[![Gitter chat](https://img.shields.io/badge/Chat-On%20Gitter-%2386f4e4?style=for-the-badge)](https://gitter.im/NiklasRosenstein/pydoc-markdown)\n\nPydoc-Markdown is a tool to create Python API documentation in Markdown format. Instead of executing your Python\ncode like so many other documentation tools, it parses it using [docspec][] instead. To run Pydoc-Markdown, you\nneed to use at least Python 3.7.\n\n[>> Go to the Documentation][Documentation]\n\n  [contrib]: https://github.com/NiklasRosenstein/pydoc-markdown/blob/develop/.github/CONTRIBUTING.md\n  [docspec]: https://niklasrosenstein.github.io/docspec/\n  [Documentation]: https://niklasrosenstein.github.io/pydoc-markdown/\n  [MkDocs]: https://www.mkdocs.org/\n  [Novella]: https://niklasrosenstein.github.io/novella/\n  [Novella build backend]: https://niklasrosenstein.github.io/pydoc-markdown/usage/novella/\n\n### Installation\n\nI recommend to install Pydoc-Markdown using Pipx.\n\n    $ pipx install pydoc-markdown[novella]\n\n### Features\n\n* Understands multiple documentation styles (Sphinx, Google, Pydoc-Markdown specific) and converts them to properly\n  formatted Markdown\n* Can parse docstrings for variables thanks to [docspec][] (`#:` block before or string literal after the statement)\n* Generates links to other API objects per the documentation syntax (e.g. `#OtherClass` for the Pydoc-Markdown style)\n\n### News\n\nStarting with __4.6.0__, development focuses on integrating with [Novella][] and use it as a replacement for\ntool-specific renderers thus far provided directly by Pydoc-Markdown (i.e. integrations with MkDocs, Hugo and\nDocusuraus). Such integrations are/will be provided by Novella instead.\n\nWith the Novella integration, you can now place generated API content in Markdown format inline with your\nexisting Markdown documentation source files using `@pydoc` tags. Check out the [Documentation][] for more\ninformation on how to use Pydoc-Markdown with Novella.\n\nThe old style of using Pydoc-Markdown with a YAML or PyProject configuration to generate files and kick off the\nbuild is now deprecated but will be maintained for the foreseeable future (at least until the next major version\nbump). It is strongly recommended to migrate your existing projects to using the Novella build backend.\n\n### Contributing to Pydoc-Markdown\n\nAll contributions are welcome! Check out the [Contribution Guidelines][contrib].\n\n### Questions / Need help?\n\nFeel free to open a topic on [GitHub Discussions](https://github.com/NiklasRosenstein/pydoc-markdown/discussions)!\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import dataclasses
import logging
import re
import typing as t

import docspec

from pydoc_markdown.interfaces import Processor, Resolver, ResolverV2
from pydoc_markdown.util.docspec import ApiSuite

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class CrossrefProcessor(Processor):
  """
  Finds references to other objects in Markdown docstrings and produces links to other
  pages. The links are provided by the current #Renderer via the #Resolver interface.

  > __Note__: This processor is a work in progress, and most of the time it just converts
  > references into inline-code.

  The syntax for cross references is as follows:

  ```
  This is a ref to another class: #PydocmdProcessor
  You can rename a ref like #this~PydocmdProcessor
  And you can append to the ref name like this: #PydocmdProcessor#s
  ```

  Renders as

  > This is a ref to another class: #PydocmdProcessor
  > You can rename a ref like #this~PydocmdProcessor
  > And you can append to the ref name like this: #PydocmdProcessor#s

  Example configuration:

  ```yaml
  processors:
    - type: crossref
  ```
  """

  #: If specified, it will be used instead of the resolver passed to #process() and the generated Markdown
  #: code for the reference uses Novella `{@link}` syntax.
  resolver_v2: t.Optional[ResolverV2] = None

  def process(self, modules: t.List[docspec.Module], resolver: t.Optional[Resolver]) -> None:
    unresolved: t.Dict[str, t.List[str]] = {}
    docspec.visit(modules, lambda x: self._preprocess_refs(x, t.cast(Resolver, resolver), ApiSuite(modules), unresolved))

    if unresolved:
      summary = []
      for uid, refs in unresolved.items():
        summary.append('  {}: {}'.format(uid, ', '.join(refs)))

      logger.warning(
        '%s cross-reference(s) could not be resolved:\n%s',
        sum(map(len, unresolved.values())),
        '\n'.join(summary),
      )

  def _preprocess_refs(
    self,
    node: docspec.ApiObject,
    resolver: t.Optional[Resolver],
    suite: ApiSuite,
    unresolved: t.Dict[str, t.List[str]],
  ) -> None:
    if not node.docstring:
      return

    def handler(match: re.Match) -> str:
      ref = match.group('ref')
      parens = match.group('parens') or ''
      trailing = (match.group('trailing') or '').lstrip('#')
      # Remove the dot from the ref if its trailing (it is probably just
      # the end of the sentence).
      has_trailing_dot = False
      if trailing and trailing.endswith('.'):
        trailing = trailing[:-1]
        has_trailing_dot = True
      elif not parens and ref.endswith('.'):
        ref = ref[:-1]
        has_trailing_dot = True

      text = ref + parens + trailing
      result: t.Optional[str] = None

      if self.resolver_v2:
        target = self.resolver_v2.resolve_reference(suite, node, ref)
        if target:
          import tomli_w
          opt = tomli_w.dumps({"text": text})
          result = f'{{@link pydoc:{".".join(x.name for x in target.path)} :with {opt}}}'

      elif resolver:
        href = resolver.resolve_ref(node, ref)
        if href:
          result = '[`{}`]({})'.format(text, href)

      if result is None:
        uid = '.'.join(x.name for x in node.path)
        unresolved.setdefault(uid, []).append(ref)
        result = '`{}`'.format(ref + parens + trailing)

      # Add back the dot.
      if has_trailing_dot:
        result += '.'

      return result

    node.docstring.content = re.sub(
      r'\B#(?P<ref>[\w\d\._]+)(?P<parens>\(\))?(?P<trailing>#[\w\d\._]+)?',
      handler,
      node.docstring.content)
